"""
Utility functions for creating flask views
"""

import logging
import traceback

from flask import current_app, render_template

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


def internal_error(exception, template_path, is_admin, db=None):
    """
    Render an "internal error" page.  The following variables will be populated when rendering the
    template:
    
    title: The page title
    message: The body of the error message to display to the user
    preformat: Boolean stating whether to wrap the error message in a pre

    As well as rendering the error message to the user, this will also log the exception

    :param exception: The exception that was caught
    :param template_path: The template to render (i.e. "main/error.html")
    :param is_admin: Can the logged in user always view detailed error reports?
    :param db: The Flask-SQLAlchemy instance

    :return: Flask Response
    """
    if db:
        try:
            db.session.rollback()
        except:
            pass

    title = str(exception)
    message = traceback.format_exc()
    preformat = True
 
    log.error('Exception caught: {}\n{}'.format(title, message))

    if current_app.config['TEST_MODE']:
        message = 'Note: You are seeing this error message because the server is in test mode.\n\n{}'.format(message)
    elif is_admin:
        message = 'Note: You are seeing this error message because you are a member of staff.\n\n{}'.format(message)
    else:
        title = '500 Internal Server Error'
        message = 'Something went wrong while processing your request.'
        preformat = False
    
    try:
        return render_template(template_path, title=title, message=message, preformat=preformat), 500
    except:
        log.exception('Error rendering error page!')
        return '500 Internal Server Error', 500



"""
Simple socket server, designed for unit testing
"""

import logging
import socketserver
import threading

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


def simple_tcp_handler(simple_server):
    class SimpleTcpHandler(socketserver.BaseRequestHandler):
        def handle(self):
            while simple_server.running:
                data = self.request.recv(1024)
                if not data:
                    log.debug('Client disconnect')
                    self.request.close()
                    break

                simple_server.add_data(data)

    return SimpleTcpHandler


class ThreadedTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    pass


class SimpleServer(object):
    def __init__(self, host='localhost', port=49932, print_data=False):
        self.host = host
        self.port = port
        self.print_data = print_data
        self.data = []
        self.data_lock = threading.Lock()
        self.server = ThreadedTCPServer((host, port), simple_tcp_handler(self))
        self.server_thread = threading.Thread(target=self.server.serve_forever)
        self.server_thread.daemon = True
        self.running = False

    def start(self):
        self.server_thread.start()
        log.info('Server tread started ({}:{})'.format(self.host, self.port))
        self.running = True

    def stop(self):
        log.info('Shutting down server')
        self.running = False
        self.server.shutdown()
        self.server.server_close()
        log.info('Server shut down complete')

    def add_data(self, value):
        with self.data_lock:
            if self.print_data:
                log.info('Received: {}'.format(value))

            self.data.append(value.decode('utf-8'))

    def get_data(self):
        with self.data_lock:
            return ''.join(self.data)

    def clear_data(self):
        with self.data_lock:
            self.data = []

    def get_and_clear_data(self):
        with self.data_lock:
            data = ''.join(self.data)
            self.data = []
            return data


if __name__ == '__main__':
    import time
    import datetime
    import signal
    import sys
    
    logging.basicConfig(level=logging.DEBUG)
    
    port = 49932
    if len(sys.argv) > 1:
        port = int(sys.argv[1])

    server = SimpleServer(port=port, print_data=True)

    def signal_handler(sig, frame):
        log.info('CTRL+C pressed - shutting down')
        server.stop()
        sys.exit(0)

    signal.signal(signal.SIGINT, signal_handler)

    server.start()
    while (True):
        log.debug('Server still running {}'.format(datetime.datetime.now()))
        data = server.get_and_clear_data()
        if data:
            log.info('Data:\n{}'.format(data))
        time.sleep(10)





"""
Functions for handling xml, normally using lxml
"""

import logging

from lxml import objectify, etree

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


def remove_namespaces(root):
    """Call this on an lxml.etree document to remove all namespaces"""
    for elem in root.getiterator():
        if not hasattr(elem.tag, 'find'):
            continue

        i = elem.tag.find('}')
        if i >= 0:
            elem.tag = elem.tag[i + 1:]

    objectify.deannotate(root, cleanup_namespaces=True)


def pretty_print_xml(xml, remove_blank_text=False):
    parser = etree.XMLParser(remove_blank_text=remove_blank_text)
    tree = etree.fromstring(xml, parser=parser)
    xml_output = etree.tostring(tree, pretty_print=True).decode()

    return xml_output



"""
This will test the connection before each query to see if the database has disconnected and recreate the connect.
Looks like magic to me...
"""

import logging

from sqlalchemy import exc
from sqlalchemy import event
from sqlalchemy.pool import Pool

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


@event.listens_for(Pool, "checkout")
def ping_connection(dbapi_connection, connection_record, connection_proxy):
    # log.debug('Pinging postgresql server...')

    cursor = dbapi_connection.cursor()
    try:
        cursor.execute("SELECT 1")
    except Exception as e:
        # optional - dispose the whole pool
        # instead of invalidating one at a time
        # connection_proxy._pool.dispose()

        # raise DisconnectionError - pool will try
        # connecting again up to three times before raising.
        log.warn('Caught dodgy postgres disconnect: %s' % e)
        raise exc.DisconnectionError()
    cursor.close()



"""
Unit tests for pagers
"""

import logging
import pytest

from littlefish.pager import SimplePager

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


class FakeQuery(object):
    def __init__(self, num_rows):
        self.list = [0] * num_rows
    
    def count(self):
        return len(self.list)

    def __getitem__(self, index):
        return self.list[index]


def test_fake_query():
    """Testing the test stuff..."""
    q = FakeQuery(12)
    
    assert q.count() == 12

    for i in range(12):
        assert q[i] == 0

    with pytest.raises(Exception):
        q[-99]
    
    with pytest.raises(Exception):
        q[12]
    
    with pytest.raises(Exception):
        q[13]
    
    with pytest.raises(Exception):
        q[44]

    assert q[3:5] == [0, 0]


def test_simple_pager():
    q1 = FakeQuery(100)
    p = SimplePager(10, 3, q1)

    assert p.total_items == 100
    assert p.total_pages == 10
    assert p.page_number == 3

    numbers = p.page_link_numbers
    assert len(numbers) == 10
    assert numbers[0] == 1
    assert numbers[9] == 10

    q2 = FakeQuery(99)
    p2 = SimplePager(10, 3, q2)
    
    assert p2.total_items == 99
    assert p2.total_pages == 10
    assert p2.page_number == 3
    numbers2 = p2.page_link_numbers
    assert len(numbers2) == 10
    assert numbers2[0] == 1
    assert numbers2[-1] == 10

    q3 = FakeQuery(101)
    p3 = SimplePager(10, 3, q3)
    
    assert p3.total_items == 101
    assert p3.total_pages == 11
    assert p3.page_number == 3
    numbers3 = p3.page_link_numbers
    assert len(numbers3) == 11
    assert numbers3[0] == 1
    assert numbers3[-1] == 11

    q4 = FakeQuery(1001)
    p4 = SimplePager(10, 3, q4)

    assert p4.total_items == 1001
    assert p4.total_pages == 101
    assert p4.page_number == 3
    numbers4 = p4.page_link_numbers
    assert len(numbers4) == 12
    assert numbers4[0] == 1
    assert numbers4[-1] == 12

    q5 = FakeQuery(1001)
    p5 = SimplePager(10, 30, q5)

    assert p5.total_items == 1001
    assert p5.total_pages == 101
    assert p5.page_number == 30
    numbers5 = p5.page_link_numbers
    assert len(numbers5) == 12
    assert numbers5[0] == 25
    assert numbers5[-1] == 36

    q6 = FakeQuery(1001)
    p6 = SimplePager(10, 98, q6, 30)

    assert p6.total_items == 1001
    assert p6.total_pages == 101
    assert p6.page_number == 98
    numbers6 = p6.page_link_numbers
    assert len(numbers6) == 30
    assert numbers6[0] == 72
    assert numbers6[-1] == 101




"""
Utility functions for dealing with files
"""

import logging
import hashlib

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


def extension_from_filename(filename):
    return filename.rsplit('.')[-1]


def read_file(filename):
    with open(filename, 'rb') as f:
        file_data = f.read()
        return file_data


def write_file(filename, data):
    with open(filename, 'wb') as f:
        f.write(data)


def file_md5sum(filename):
    """
    :param filename: The filename of the file to process
    :returns: The MD5 hash of the file
    """
    hash_md5 = hashlib.md5()
    with open(filename, 'rb') as f:
        for chunk in iter(lambda: f.read(1024 * 4), b''):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def bytes_md5sum(bytes_):
    hash_md5 = hashlib.md5()
    hash_md5.update(bytes_)
    return hash_md5.hexdigest()




"""
This contains functions to manipulate and display timestamps
"""

import logging
import datetime
import calendar

import dateutil.parser
import pytz

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)

utc = pytz.utc
local = pytz.timezone('Europe/London')


def to_local_time(time_in):
    if not hasattr(time_in, 'tzinfo'):
        # This is a date with no timestamp info so we can't do the conversion
        return time_in
    elif time_in.tzinfo is None:
        # If there is no timezone, localise it to UTC
        time_utc = utc.localize(time_in)
    elif hasattr(time_in.tzinfo, 'zone') and time_in.tzinfo.zone == 'UTC':
        # Already UTC - no need to localize
        time_utc = time_in
    elif hasattr(time_in.tzinfo, 'tzname') and time_in.tzinfo.tzname(time_in) == 'UTC':
        # Already UTC
        time_utc = time_in
    else:
        # Not a UTC timestamp
        raise Exception('Not a UTC time: %s' % time_in)
    
    time_local = time_utc.astimezone(local)
    # Strip off timezone info
    return time_local.replace(tzinfo=None)


def get_local_time():
    return to_local_time(datetime.datetime.utcnow())


def date_to_local_time(date):
    midnight = datetime.time()
    return local.localize(datetime.datetime.combine(date, midnight))


def date_to_datetime(date):
    return datetime.datetime(date.year, date.month, date.day)


def to_utc_time(time_in):
    if not hasattr(time_in, 'tzinfo'):
        # This is a date with no timestamp info so we can't do the conversion
        return time_in
    elif time_in.tzinfo is None:
        # If there is no timezone, localise it to local time
        time_local = local.localize(time_in)
    elif time_in.tzinfo.zone == local.zone:
        # Already local - no need to localize
        time_local = time_in
    else:
        # Not a local timestamp
        raise Exception('Not a local time: %s' % time_in)

    time_utc = time_local.astimezone(utc)
    # Strip off timezone info
    return time_utc.replace(tzinfo=None)


def format_datetime(datetime, convert_to_local=True):
    if datetime:
        if convert_to_local:
            local_datetime = to_local_time(datetime)
            return local_datetime.strftime('%d/%m/%Y %H:%M')
        else:
            return datetime.strftime('%d/%m/%Y %H:%M')
    else:
        return ''


def format_datetime_seconds(datetime, convert_to_local=True):
    if datetime:
        if convert_to_local:
            local_datetime = to_local_time(datetime)
            return local_datetime.strftime('%d/%m/%Y %H:%M:%S')
        else:
            return datetime.strftime('%d/%m/%Y %H:%M:%S')
    else:
        return ''


def format_datetime_long(datetime, convert_to_local=True):
    if datetime:
        if convert_to_local:
            local_datetime = to_local_time(datetime)
            return local_datetime.strftime('%A %d %B %Y %H:%M')
        else:
            return datetime.strftime('%A %d %B %Y %H:%M')
    else:
        return ''


def format_datetime_long_seconds(datetime, convert_to_local=True):
    if datetime:
        if convert_to_local:
            local_datetime = to_local_time(datetime)
            return local_datetime.strftime('%A %d %B %Y %H:%M:%S')
        else:
            return datetime.strftime('%A %d %B %Y %H:%M:%S')
    else:
        return ''


def format_date(datetime, convert_to_local=True):
    if datetime:
        try:
            if convert_to_local:
                local_datetime = to_local_time(datetime)
                return local_datetime.strftime('%d/%m/%Y')
            else:
                return datetime.strftime('%d/%m/%Y')
        except ValueError:
            log.warning('Invalid date: %s' % datetime)
            return '????'
    else:
        return ''


def format_date_long(datetime, convert_to_local=True):
    if datetime:
        try:
            if convert_to_local:
                local_datetime = to_local_time(datetime)
                return local_datetime.strftime('%A %-d %B %Y')
            else:
                return datetime.strftime('%A %-d %B %Y')
        except ValueError:
            log.warning('Invalid date: %s' % datetime)
            return '????'
    else:
        return ''


def format_date_long_no_day(datetime, convert_to_local=True):
    if datetime:
        try:
            if convert_to_local:
                local_datetime = to_local_time(datetime)
                return local_datetime.strftime('%-d %B %Y')
            else:
                return datetime.strftime('%-d %B %Y')
        except ValueError:
            log.warning('Invalid date: %s' % datetime)
            return '????'
    else:
        return ''


def format_time(datetime, convert_to_local=True):
    if datetime:
        if convert_to_local:
            local_datetime = to_local_time(datetime)
            return local_datetime.strftime('%H:%M:%S')
        else:
            return datetime.strftime('%H:%M:%S')
    else:
        return ''


def format_time_delta(delta):
    if delta.days:
        return str(delta.days) + ' days ' + format_duration_seconds(delta.seconds)
    return format_duration_seconds(delta.seconds)


def format_duration_seconds(seconds):
    h = seconds // 3600
    seconds = seconds % 3600
    m = seconds // 60
    # I'm sure there is a better way of doing this!
    if m >= 10:
        return "%i:%i" % (h, m)
    else:
        return "%i:0%i" % (h, m)


def add_working_days(num_days, date=None, include_saturday=False):
    if date is None:
        date = datetime.date.today()
        
    for i in range(num_days):
        date += datetime.timedelta(days=1)
        while date.weekday() == 6 or (not include_saturday and date.weekday() == 5):
            date += datetime.timedelta(days=1)
        
    return date


def datetime_from_datepicker(date_string):
    """Turn a string from a jQuery UI datetpicker into a datetime object"""
    return datetime.datetime.strptime(date_string, '%d/%m/%Y')


def date_from_datepicker(date_string):
    """Turn a string from a jQuery UI datetpicker into a datetime object"""
    return datetime_from_datepicker(date_string).date()


def datetime_to_datepicker(timestamp):
    """Turn a datetime object back into a string for a JQuery UI datepicker"""
    return timestamp.strftime('%d/%m/%Y')


def add_months(months, timestamp=datetime.datetime.utcnow()):
    """Add a number of months to a timestamp"""
    month = timestamp.month
    new_month = month + months
    years = 0
    
    while new_month < 1:
        new_month += 12
        years -= 1
    
    while new_month > 12:
        new_month -= 12
        years += 1
    
    # month = timestamp.month
    year = timestamp.year + years

    try:
        return datetime.datetime(year, new_month, timestamp.day, timestamp.hour, timestamp.minute, timestamp.second)
    except ValueError:
        # This means that the day exceeds the last day of the month, i.e. it is 30th March, and we are finding the day
        # 1 month ago, and it is trying to return 30th February
        if months > 0:
            # We are adding, so use the first day of the next month
            new_month += 1
            if new_month > 12:
                new_month -= 12
                year += 1
            
            return datetime.datetime(year, new_month, 1, timestamp.hour, timestamp.minute, timestamp.second)
        else:
            # We are subtracting - use the last day of the same month
            new_day = calendar.monthrange(year, new_month)[1]
            return datetime.datetime(year, new_month, new_day, timestamp.hour, timestamp.minute, timestamp.second)


def add_months_to_date(months, date):
    """Add a number of months to a date"""
    month = date.month
    new_month = month + months
    years = 0

    while new_month < 1:
        new_month += 12
        years -= 1

    while new_month > 12:
        new_month -= 12
        years += 1

    # month = timestamp.month
    year = date.year + years

    try:
        return datetime.date(year, new_month, date.day)
    except ValueError:
        # This means that the day exceeds the last day of the month, i.e. it is 30th March, and we are finding the day
        # 1 month ago, and it is trying to return 30th February
        if months > 0:
            # We are adding, so use the first day of the next month
            new_month += 1
            if new_month > 12:
                new_month -= 12
                year += 1

            return datetime.datetime(year, new_month, 1)
        else:
            # We are subtracting - use the last day of the same month
            new_day = calendar.monthrange(year, new_month)[1]
            return datetime.datetime(year, new_month, new_day)


def unix_time(dt=None, as_int=False):
    """Generate a unix style timestamp (in seconds)"""
    if dt is None:
        dt = datetime.datetime.utcnow()

    if type(dt) is datetime.date:
        dt = date_to_datetime(dt)

    epoch = datetime.datetime.utcfromtimestamp(0)
    delta = dt - epoch
    
    if as_int:
        return int(delta.total_seconds())

    return delta.total_seconds()


def current_time_millis(dt=None):
    unix_time_seconds = unix_time(dt=dt)
    return int(unix_time() * 1000)


def format_datetime_iso8601(datetime):
    return datetime.strftime('%Y-%m-%dT%H:%M:%SZ')


def datetime_from_iso8601(string):
    return dateutil.parser.parse(string)


def is_christmas_period():
    """Is this the christmas period?"""
    now = datetime.date.today()
    if now.month != 12:
        return False
    if now.day < 15:
        return False
    if now.day > 27:
        return False
    return True


def get_end_of_day(timestamp):
    """
    Given a date or a datetime, return a datetime at 23:59:59 on that day
    """
    return datetime.datetime(timestamp.year, timestamp.month, timestamp.day, 23, 59, 59)




"""
Code to send email messages via an SMTP server
"""

import logging
from email.mime.text import MIMEText
import smtplib
import json
import traceback
import pprint
import email.utils

from . import timetool

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)

host = None
port = None
username = None
password = None
use_tls = None
default_email_from = None
email_to_override = None
dump_email_body = None
_configured = False
_error_reporting_obscured_fields = None


def init(app):
    global host, port, username, password, use_tls, default_email_from, email_to_override,\
        dump_email_body, _configured

    if _configured and not app.config['TESTING']:
        raise Exception('Multiple calls to {}.init(app)'.format(__name__))

    host = app.config['SMTP_HOST']
    port = app.config['SMTP_PORT']
    username = app.config['SMTP_USERNAME']
    password = app.config['SMTP_PASSWORD']
    use_tls = app.config['SMTP_USE_TLS']

    log.info('LFS Mailer using {}@{}:{}{}'.format(username, host, port, ' (TLS)' if use_tls else ''))

    default_email_from = app.config['DEFAULT_EMAIL_FROM']

    # If this is set, we will send all emails here
    email_to_override = app.config['EMAIL_TO_OVERRIDE']
    # Whether or not to dump the email body
    dump_email_body = app.config['DUMP_EMAIL_BODY']

    _configured = True


def format_address(email_address, name=None):
    if name is None:
        return email_address
    return '{} <{}>'.format(name, email_address)


def send_text_mail_single(to_email_address, to_name, subject, body, from_address=None):
    to = format_address(to_email_address, to_name)

    send_text_mail([to], subject, body, from_address)


def send_text_mail(recipient_list, subject, body, from_address=None):
    send_mail(recipient_list, subject, body, html=False, from_address=from_address)


def send_html_mail_single(to_email_address, to_name, subject, body, from_address=None):
    to = format_address(to_email_address, to_name)

    send_html_mail([to], subject, body, from_address)


def send_html_mail(recipient_list, subject, body, from_address=None):
    send_mail(recipient_list, subject, body, html=True, from_address=from_address)


def send_mail(recipient_list, subject, body, html=False, from_address=None):
    """
    :param recipient_list: List of recipients i.e. ['testing@fig14.com', 'Stephen Brown <steve@fig14.com>']
    :param subject: The subject
    :param body: The email body
    :param html: Is this a html email? Defaults to False
    :param from_address: From email address or name and address i.e. 'Test System <errors@test.com>
    :return:
    """
    if not _configured:
        raise Exception('LFS Mailer hasn\'t been configured')

    if from_address is None:
        from_address = default_email_from
    
    mime_type = 'html' if html else 'plain'
    log.debug('Sending {} mail to {}: {}'.format(mime_type, ', '.join(recipient_list), subject))
    if dump_email_body:
        log.info(body)

    s = smtplib.SMTP(host, port)

    if use_tls:
        s.ehlo()
        s.starttls()
        s.ehlo()
    
    if username:
        s.login(username, password)

    if email_to_override:
        subject = '[to %s] %s' % (', '.join(recipient_list), subject)
        recipient_list = [email_to_override]
        log.info('Using email override: %s' % ', '.join(recipient_list))

    msg = MIMEText(body, mime_type, 'utf-8')
    msg['To'] = ', '.join(recipient_list)
    msg['Subject'] = subject
    msg['From'] = from_address
    msg['Date'] = email.utils.formatdate()

    s.sendmail(from_address, recipient_list, msg.as_string())
    s.quit()


class LfsSmtpHandler(logging.Handler):
    """
    A handler class which sends an SMTP email for each logging event.  This has been customised to (easily) work with
    lfsmailer
    """
    def __init__(self, fromaddr, toaddrs, subject, max_sends_per_minute=15):
        """
        Initialize the handler.

        Initialize the instance with the from and to addresses and subject
        line of the email. To specify a non-standard SMTP port, use the
        (host, port) tuple format for the mailhost argument. To specify
        authentication credentials, supply a (username, password) tuple
        for the credentials argument. To specify the use of a secure
        protocol (TLS), pass in a tuple for the secure argument. This will
        only be used when authentication credentials are supplied. The tuple
        will be either an empty tuple, or a single-value tuple with the name
        of a keyfile, or a 2-value tuple with the names of the keyfile and
        certificate file. (This tuple is passed to the `starttls` method).
        """
        super(LfsSmtpHandler, self).__init__()
        self.fromaddr = fromaddr
        if isinstance(toaddrs, str):
            toaddrs = [toaddrs]
        self.toaddrs = toaddrs
        self.subject = subject
        self._timeout = 5.0
        self.max_sends_per_minute = max_sends_per_minute
        self.rate_limiter = []

        # Default formatter
        self.setFormatter(logging.Formatter('''
Message type:       %(levelname)s
Location:           %(pathname)s:%(lineno)d
Module:             %(module)s
Function:           %(funcName)s
Time:               %(asctime)s

Message:

%(message)s
'''))
    
    def add_details(self, message):
        """
        Add extra details to the message.  Separate so that it can be overridden
        """
        msg = message
        # Try to append Flask request details
        try:
            from flask import request
            url = request.url
            method = request.method
            endpoint = request.endpoint

            # Obscure password field and prettify a little bit
            form_dict = dict(request.form)
            for key in form_dict:
                if key.lower() in _error_reporting_obscured_fields:
                    form_dict[key] = '******'
                elif len(form_dict[key]) == 1:
                    form_dict[key] = form_dict[key][0]

            form = pprint.pformat(form_dict).replace('\n', '\n          ')

            msg = '%s\nRequest:\n\nurl:      %s\nmethod:   %s\nendpoint: %s\nform:     %s\n' % \
                (msg, url, method, endpoint, form)
        except:
            traceback.print_exc()

        # Try to append the session
        try:
            from flask import session
            from flask.json import JSONEncoder
            session_str = json.dumps(
                dict(**session),
                indent=2,
                cls=JSONEncoder
            )
            msg = '%s\nSession:\n\n%s\n' % (msg, session_str)
        except:
            traceback.print_exc()
        
        return msg

    def emit(self, record):
        """
        Emit a record.

        Format the record and send it to the specified addressees.
        """
        try:
            # First, remove all records from the rate limiter list that are over a minute old
            now = timetool.unix_time()
            one_minute_ago = now - 60
            new_rate_limiter = [x for x in self.rate_limiter if x > one_minute_ago]
            log.debug('Rate limiter %s -> %s' % (len(self.rate_limiter), len(new_rate_limiter)))
            self.rate_limiter = new_rate_limiter

            # Now, get the number of emails sent in the last minute.  If it's less than the threshold, add another
            # entry to the rate limiter list
            recent_sends = len(self.rate_limiter)
            send_email = recent_sends < self.max_sends_per_minute
            if send_email:
                self.rate_limiter.append(now)

            msg = self.format(record)
            msg = self.add_details(msg)

            # Finally send the message!
            if send_email:
                send_text_mail(self.toaddrs, self.subject, msg, self.fromaddr)
            else:
                log.info('!! WARNING: Not sending email as too many emails have been sent in the past minute !!')
                log.info(msg)

        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)


def init_error_emails(send_error_emails, send_warning_emails, from_address, to_addresses, subject,
                      logger=None, obscured_fields=['password']):
    global _error_reporting_obscured_fields
    
    _error_reporting_obscured_fields = obscured_fields

    if send_error_emails or send_warning_emails:
        log.info('Setting up error / warning emails')
        error_handler = LfsSmtpHandler(from_address, to_addresses, subject)

        if send_warning_emails:
            log.info('Sending WARNING emails as well as ERRORs')
            error_handler.setLevel(logging.WARNING)
        else:
            log.info('Only sending ERROR emails')
            error_handler.setLevel(logging.ERROR)
        
        if logger is None:
            logger = logging.getLogger()

        logger.addHandler(error_handler)



"""
For converting colours between rgb and hsl
"""

import logging
from math import floor

__author__ = 'Yu Lee Paul (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


def rgb_to_hsl(r, g, b):
    """
    Converts an RGB color value to HSL.
    :param r: The red color value
    :param g: The green color value
    :param b: The blue color value
    :return: The HSL representation
    """
    r = float(r) / 255.0
    g = float(g) / 255.0
    b = float(b) / 255.0

    max_value = max(r, g, b)
    min_value = min(r, g, b)

    h = None
    s = None
    l = (max_value + min_value) / 2
    d = max_value - min_value

    if d == 0:
        # achromatic
        h = 0
        s = 0
    else:
        s = d / (1 - abs(2 * l - 1))

        if r == max_value:
            h = 60 * ((g - b) % 6)
            if b > g:
                h += 360
        if g == max_value:
            h = 60 * ((b - r) / d + 2)
        if b == max_value:
            h = 60 * ((r - g) / d + 4)

    return round(h, 2), round(s, 2), round(l, 2)


def hsl_to_rgb(h, s, l):
    h = float(h)
    s = float(s)
    l = float(l)

    c = (1 - abs(2 * l - 1)) * s
    mod = (h / 60.0) % 2
    x = c * (1 - abs(mod - 1))
    m = l - (c / 2)

    if h < 60:
        r = c
        g = x
        b = 0
    elif h < 120:
        r = x
        g = c
        b = 0
    elif h < 180:
        r = 0
        g = c
        b = x
    elif h < 240:
        r = 0
        g = x
        b = c
    elif h < 300:
        r = x
        g = 0
        b = c
    else:
        r = c
        g = 0
        b = x

    r = (r + m) * 255
    g = (g + m) * 255
    b = (b + m) * 255

    return floor(r), floor(g), floor(b)


def html_color_to_rgba(html_colour, alpha):
    """
    :param html_colour: Colour string like FF0088
    :param alpha: Alpha value (opacity)
    :return: RGBA semitransparent version of colour for use in css
    """
    html_colour = html_colour.upper()
    if html_colour[0] == '#':
        html_colour = html_colour[1:]

    r_str = html_colour[0:2]
    g_str = html_colour[2:4]
    b_str = html_colour[4:6]

    r = int(r_str, 16)
    g = int(g_str, 16)
    b = int(b_str, 16)

    return 'rgba(%s, %s, %s, %s)' % (r, g, b, alpha)


def blend_html_colour_to_white(html_colour, alpha):
    """
    :param html_colour: Colour string like FF552B or #334455
    :param alpha: Alpha value
    :return: Html colour alpha blended onto white
    """
    html_colour = html_colour.upper()
    has_hash = False
    if html_colour[0] == '#':
        has_hash = True
        html_colour = html_colour[1:]

    r_str = html_colour[0:2]
    g_str = html_colour[2:4]
    b_str = html_colour[4:6]

    r = int(r_str, 16)
    g = int(g_str, 16)
    b = int(b_str, 16)

    r = int(alpha * r + (1 - alpha) * 255)
    g = int(alpha * g + (1 - alpha) * 255)
    b = int(alpha * b + (1 - alpha) * 255)
    
    out = '{:02X}{:02X}{:02X}'.format(r, g, b)
    if has_hash:
        out = '#' + out

    return out


